// LifePo4 Cell balancer 
// external reference on pin 5 Aref (2.048V)
// measured Ubat on pin 3 throught divider (Ubat/2)
// sends respone (data) over serial port Pin 6 (TX)
// receives commands and requests over serial receive port Pin 2 (RX)
// read Ubat and temperature and send it to serial (Attiny85 Pin 6 (TX))
// Serial uses inverted logic levels !!! 0 - low, 1 - High
// send data format  A
// serial comunication format
// A (target address 0..F are cells, ( - data for balancer manager
// command:
//		QUA get actual Ubat in mV, response (AUA9999
//		QTA get actual temperature in kelvin, response (ATA999
//		QBS get starting balance voltage in mV, response (ABS9999
//		QBH get balance histerezis voltage in mV, response (ABH99
//		QBT get balance histerezis time in seconds, response (ABT99
// set commands
//		SDB - start dynamic balancing, response (ASDBOK
//		EDB - end dynamic balancing, response (AEDBOK
//		BSV9999 - set start balance voltage, response (ABSVOK
//		BHV99 - set hysterezis in mV (range 0 ... 99mV), response (ABHVOK
//		BHT99 - set balance hysterezis time (range 0..99) in seconds, response (ABHTOK
//		RFC - set reference corection (-10..10), response (ARFCOK
//
// all commands and responses must end with correct <CRC><cr>
//uknow commnad is discarded, correct commands and reponses are forwarder from RX to TX without modification

// version 0 from 13.1.2019



#include "ADConversion.h"
#include "Buffer.h"
#include <avr\sleep.h>
#include <ATTinyCore.h>
#include <SoftwareSerial.h>
#include <Arduino.h>

//inverted logic levels, RX pin 3, TX pin 6
// low 0V, high Vcc
SoftwareSerial swserial(3, 1, true); 

//balancer ovladany pinom 7  
#define BALANACER_OUTPUT PB2

short adc2_gain_corr = 0;
short adc2_offset_corr = 0;
bool adc2_raw = false;

bool balance_on = false;

BufferClass rx_data;
bool buff_tx_hasadata = false;

bool vcc_on = false;
bool adc2_on = true;
bool temp_on = false;
// true ak bola poziadavka na zaslanie udajov
bool sendData = false;

// vysledok posledneho merania
int adc2_result = 0;

// Variables for the Sleep/power down modes:
volatile boolean f_wdt = 1;

ADConversion adc;

// the setup function runs once when you press reset or power the board
void setup() {
  swserial.begin(4800);
  pinMode(BALANACER_OUTPUT, OUTPUT);
  digitalWrite(BALANACER_OUTPUT, LOW);
  //set watch dog to 4s
  setup_watchdog(8); 
}

// Watchdog Interrupt Service / is executed when watchdog timed out
ISR(WDT_vect) {
	f_wdt = 1;  // set global flag
}

// the loop function runs over and over again forever
void loop() 
{
	if (f_wdt == 1) {  // wait for timed out watchdog / flag is set when a watchdog timeout occurs
		f_wdt = 0;       // reset flag, othervise processor will reset

		// flag to send data on timeout
		sendData = true;
		//start ADC
		adc.start();
	}
	swserial.listen(); //skusal som v setup a blblo
	while (swserial.available())
	{
		if (rx_data.isFull())
		{
			//start from begin of buffer (clear)
			rx_data.clear();
		}

		char chr = swserial.read();
		// if the incoming character is a newline, set a flag so the main loop can
		// do something about it:
		if ((chr == '\r') || (chr == '\n'))  
		{
			if (rx_data.len > 0)
				buff_tx_hasadata = true;
		}
		else
		{
			rx_data.addChar(chr);
		}
	}

	if (buff_tx_hasadata)
	{
    swserial.println( rx_data.data);
		//process data in input buffer
		if (strncmp(rx_data.data, "ADC2", 4) == 0)
		{
			adc2_on = !adc2_on;
			// clear buffer
			rx_data.clear();
			buff_tx_hasadata = false;
		}
		//else
		//if (strncmp(rx_data.data, "TMP", 3) == 0)
		//{
		//	temp_on = !temp_on;
		//	// clear buffer
		//	rx_data.clear();
		//	buff_tx_hasadata = false;
		//}
		//else
		//if (strncmp(rx_data.data, "AD2GC", 5) == 0)
		//{
		//	//set vcc_gain_corr
		//	short corr = (short)strtol((rx_data.data + 5), nullptr, 10);
		//	//check for range error
		//	if ((corr >= 512) || (corr <= -512))
		//		corr = 0;

		//	//store gain correction
		//	adc2_gain_corr = corr;
		//	// clear buffer
		//	rx_data.clear();
		//	buff_tx_hasadata = false;
		//}
		else
		if (strncmp(rx_data.data, "AD2OC", 5) == 0)
		{
			//set vcc_offset_corr
			short corr = (short)strtol((rx_data.data + 5), nullptr, 10);
			//check for range error
			if ((corr >= 512) || (corr <= -512))
				corr = 0;

			//store offset correction
			adc2_offset_corr = corr;
			// clear buffer
			rx_data.clear();
			buff_tx_hasadata = false;
		}
		else
		if (strncmp(rx_data.data, "AD2RAW", 6) == 0)
		{
			//adc RAW or recalculated
			adc2_raw = !adc2_raw;
			// clear buffer
			rx_data.clear();
			buff_tx_hasadata = false;
		}
		else
		//if (strncmp(rx_data.data, "AD2REF", 6) == 0)
		//{
		//	//adc RAW or recalculated
		//	adc2_ref = !adc2_ref;
		//	// clear buffer
		//	rx_data.clear();
		//	buff_tx_hasadata = false;
		//}
		//else
		if (strncmp(rx_data.data, "BAL", 3) == 0)
		{
			//switch on or off balancer
			balance_on = !balance_on;
			// clear buffer
			rx_data.clear();
			buff_tx_hasadata = false;
			digitalWrite(BALANACER_OUTPUT, balance_on);
		}
		else
		{
			//uknow command - reset buffer
			// clear buffer
			rx_data.clear();
			buff_tx_hasadata = false;
		}
	}

	if (!adc.isConverting())
	{
		//adc is finished or stopped
		int adcVal = adc.getMeasuredValue();
		if (adcVal >= 0)
		{
			//finished measurement
			//pouzivam externu refernciu 2042mV, delic 1:1
			if (!adc2_raw)
			{
				//result = adc_val * Uref/1024*2= result * 2042L*2/1024
				adc2_result = adcVal * (4084L - adc2_offset_corr) / 1024L; //(result * (2600L - adc2_gain_corr))/1000L;
			}
			else
				adc2_result = adcVal;

			//

		}
		else
		{
			//adc stopped
		}
	}

	if (sendData)
	{
		//send data
		if (adc2_on)
			swserial.print(adc2_result);
		else
			swserial.print(F("off"));

		//swserial.write(0x20); //space

		//if (temp_on)
		//	swserial.print(temperature);
		//else
		//	swserial.print(F("off"));

		swserial.write(0x0D); //CR
		swserial.write(0x0A); //LF 

		sendData = false;
	}

	if ((!sendData) && (!adc.isConverting()))
	{
		//sleep and wait for Watchdog or serial interrupt
		system_sleep();  // Set the unit to sleep
	}
}


// 0=16ms, 1=32ms,2=64ms,3=128ms,4=250ms,5=500ms
// 6=1 sec,7=2 sec, 8=4 sec, 9= 8sec
void setup_watchdog(int ii) {

	byte bb;
	int ww;
	if (ii > 9) ii = 9;
	bb = ii & 7;
	if (ii > 7) bb |= (1 << 5);
	bb |= (1 << WDCE);
	ww = bb;
	//always clear the WDRF flag and the WDE control bit in the initialization routine
	MCUSR &= ~(1 << WDRF);
	// start timed sequence
	WDTCR |= (1 << WDCE) | (1 << WDE);
	// set new watchdog timeout value
	WDTCR = bb;
	WDTCR |= _BV(WDIE);
}

// set system into the sleep state 
// system wakes up when watchdog is timed out
void system_sleep() {

	set_sleep_mode(SLEEP_MODE_PWR_DOWN); // sleep mode is set here
	sleep_enable();

	sleep_mode();                        // System actually sleeps here

	sleep_disable();                     // System continues execution here when watchdog timed out 
}

/* end of file */
